#ifndef HARKTOOL_H
#define HARKTOOL_H
#ifndef WIN32
#define _GNU_SOURCE
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <sndfile.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <limits.h>
#include <stdbool.h>
#include <popt.h>

#include <sys/types.h>
#include <sys/stat.h>

#include "libharkio3.h"
//#include <libxml/xmlreader.h>



//#define  ARGV_LEN     1024     // argv 最大長
//#define  PATH_LEN     1024     // file path 最大長
// #define  HARKTOOL_MSG_LEN      1024     // message 最大長
#define  FILE_PREFIX  "float"  //
#define  DIRECT_SOUND_POSITION_LIMIT  200
#define  SOUND_SPEED  343
#define  MAX_CH       64
#define  two_pi  (2.0f * pi)

#define X_AXIS 0
#define Y_AXIS 1
#define Z_AXIS 2


//#ifdef __unix__
//#define COMPLEX_FLOAT  __complex__ float
//#define REAL  __real__
//#define IMAG  __imag__
#define COMPLEX_FLOAT        float _Complex
#define COMPLEX_DOUBLE       double _Complex
#define COMPLEX_FLOAT_SIZE   sizeof(float _Complex)
//windows
//#else
//#define COMPLEX_FLOAT        complex<float>
//#define COMPLEX_DOUBLE       complex<double>
//#endif





// #define LOG_I       (0)     // インフォメーション
// #define LOG_D       (1)     // インフォメーション
// #define LOG_W       (2)     // ワーニング
// #define LOG_E       (3)     // エラー

// #define harktool_LogI(...)  Log(LOG_I,  __func__, __LINE__, __VA_ARGS__)
// #define harktool_LogD(...)  Log(LOG_D,  __func__, __LINE__, __VA_ARGS__)
// #define harktool_LogW(...)  Log(LOG_W,  __func__, __LINE__, __VA_ARGS__)
// #define harktool_LogE(...)  Log(LOG_E,  __func__, __LINE__, __VA_ARGS__)


#define harktool_LogI(...)	harkio_Log_printf(harkio_Log_Info,  __VA_ARGS__)
#define harktool_LogD(...)	harkio_Log_printf(harkio_Log_Debug,  __VA_ARGS__)
#define harktool_LogW(...)	harkio_Log_printf(harkio_Log_Warning,  __VA_ARGS__)
#define harktool_LogE(...)	harkio_Log_printf(harkio_Log_Error,  __VA_ARGS__)


enum DimType        { FLOAT, DOUBLE , COMPLEX_ARRAY };
enum TFMethod       { MUSIC, GSS };
enum TFType         { LOC , SEP , LOC_SEP};
enum NormalizeType  { normalize_src, normalize_mic, normalize_freq };
enum MatrixType     { TF_VECT  ,   SEP_TF_VECT, TMP_TF_VECT , IR_VECT   , TMP_IR_VECT , WAV_DATA ,
                      MIC_POS_VECT, SRC_POS_VECT, FLT_PATH  , TSP_WAV_NAME ,IR_WAV_NAME ,
                      IR_WAV_PATH , SORT_POS_ID , MIC_POS_ID, IR_POS_ID ,
                      BUFFER      , BUFF_A      , BUFF_a    , BUFF_w
                    };
enum PositionType   { MIC , TSP , IMPULSE , NOISE , SRC };
enum ExitType       { SUCCESS  ,  FAILURE };
enum DirType        { DIR_TYPE , NON_DIR_TYPE };
enum FlagType       { ENABLE ,  DISABLE };

enum CopyType       { TO_BACK ,  TO_IRvect };

enum SoundType      { RAW, WAV };

static const double pi      = 3.14159265358979323846;






// ログを出力する。
// extern void Log(int logType, const char *fileName, int lineNo, char *format, ...);


#define BUFFSIZE 128



//struct ARGV {
//    int          tf_type                ; // 引数 伝達関数種別 MUSIC /GSS (output_type)
//    int          direct_sound_length    ; //
//    char         mic_file[PATH_LEN]     ; // mic file (引数)
//    char         tsp_file[PATH_LEN]     ; // TSP List file (引数)
//    char         impulse_file[PATH_LEN] ; // ImpulseList file      (引数)
//    char         noise_file[PATH_LEN]   ; // noise file (引数)
//    char         input[PATH_LEN]        ; // 引数 ( tsp001.wav)
//    char         output[PATH_LEN]       ; // 引数 作成する伝達関数のzip (hark.zip)
//    int          normalize_src          ; // 正規化(SRC) 引数
//    int          normalize_mic          ; // 正規化(mic) 引数
//    int          normalize_freq         ; // 正規化(freq)引数
//    char         loglevel[10]           ; // ログ LEVEL
//    int          start                  ; //  mouthTSP only
//    int          end                    ; //  mouthTSP only
//    int          mch                    ; //  mouthTSP only
//    int          nch                    ; //  mouthTSP only
//    int          fltlen                 ; // Number of samples output as TF [Integer]
//    int          hop                    ; // mouthTSP only :Shift length for TF calculation [Integer]
//    int          headmargin             ; // mouthTSP only :Head margin of impulse responces in samples [Integer(Default:0)]
//    int          enable_wn              ; // mouthTSP only :Enables adding white noise to escape unstable zero signal [Bool]
//    double       wn_level               ; // mouthTSP only :Level of white noise. Set enable_wn=true if you want to use this. [Double]
//    int          enable_normalize       ; // mouthTSP only :Enable normalizing output TF
//
//}  ;



struct TMP_PATH {
    char         date_time[BUFFSIZE]    ; // ディレクトリ名作成用現在日時分秒
    char         tsp_path[BUFFSIZE]     ; // tspファイル格納パス   TSP storage path
    char         ir_path[BUFFSIZE]      ; // irファイル格納パス
    char         tf_path[BUFFSIZE]      ; // tfファイル格納パス
    char**       flt_path               ; // CalcSysimpで作成したインパルス応答ファイル格納path名(Nsrc分)
    char**       tsp_wav_name           ; // TSPファイル名(Nsrc分)
    char         outimpfilehead[PATH_MAX];// CalcSysyImpで使用するpath+header部
};//tmp


struct WAVE {
    long         file_size;
    int          total_frame;
    float**      data;

};//wav






struct GLOBAL {
    int              Nsrc                   ; // 音源数
    int              Nmic                   ; // mic数
    int              Nfft                   ; // fft数 // harkio_Config

    enum PositionType PosType               ; // position type(tsp/impulse)
    int              SyncAverage            ; // harkio_Config_get
    int              TspOffset              ; // harkio_Config_get
    int              PeakSearchFrom         ; // harkio_Config_get
    int              PeakSearchTo           ; // harkio_Config_get
    int              SamplingRate           ; // harkio_Config_get
    int              SignalMax              ; // harkio_Config_get
    int              TspLength              ; // harkio_Config_get
    char             *OrgTspPath            ; // harkio_Config_get //TSP元音源パス （16384.little_endian.wav）

    COMPLEX_FLOAT*** TFvect                 ; // 伝達関数領域
    COMPLEX_FLOAT*** LocTFvect              ; // 定位用伝達関数領域(定位,分離両方同時作成時のみ使用)
    COMPLEX_FLOAT*** SepTFvect              ; // 分離用伝達関数領域(定位,分離両方同時作成時のみ使用)
    COMPLEX_FLOAT*** TmpTFvect              ; // 伝達関数 作業領域
    int              TFsize                 ; // 伝達関数領域のsize(Nsrc * Nmic * ( Nfft/2 + 1) )

    COMPLEX_FLOAT*   buff_a                 ; // fft前のデータ(計算用領域)
    COMPLEX_FLOAT*   buff_A                 ; // fft後のデータ(計算用領域)

//    float***         IRvect                 ; // インパルス応答格納領域
    float***         LocIRvect              ; // 定位用インパルス応答格納領域
    float***         SepIRvect              ; // 分離用インパルス応答格納領域
    float***         TmpIRvect              ; // インパルス応答 計算用TEMP領域
//    float***         BackIRvect             ;  // インパルス応答 Backup領域
    int              IRsize                 ; // インパルス応答格納領域のsize
    //                                        //  npos(72) x Nmic(8) x Nfft(512) 分の領域確保

    float**          MicPosVect             ; // mic position 格納領域
    float**          SrcPosVect             ; // src position 格納領域
    float            PosMean[3]             ; // 平均値

    harkio_Matrix**  Loctf                  ; // 定位伝達関数格納
    harkio_Matrix**  Septf                  ; // 分離伝達関数格納

    harkio_TransferFunction *TF             ; // 伝達関数

    harkio_XML*      mic_xml                ; // 伝達関数作成時のマイク位置
    harkio_XML*      src_xml                ; // 伝達関数作成時の音源位置
    harkio_XML*      noise_xml              ; // 伝達関数作成時のノイズ位置


    int*             MicPosId               ; // 全position idをソートした結果を格納
    int*             SrcPosId                ; // 全position idをソートした結果を格納


    struct TMP_PATH  tmp  ;

    struct WAVE      wav   ;

    char* readmeMsg;

}; //g

/**
 * xmlのpositionに追加する情報
 */
typedef struct {
    float tfOffset;
    int cutFrom;
    int cutTo;
} POSITION_INFO;

typedef struct {
    POSITION_INFO loc;
    POSITION_INFO sep;
} SRC_POS_INFO;

int harktoolcli_calctfimp(int argc, char *argv[]);
int harktoolcli_calctfimpExec(const char* impFile, const char* micFile,
        enum TFType outputType, const char* outputFile, int directLength,
        int reverbLength, int normSrc, int normMic, int normFreq, bool resetMic);
int harktoolcli_calctfrec(int argc, char *argv[]);
int harktoolcli_calctfrecExec(const char* tspFile, const char* micFile,
        enum TFType outputType, const char* outputFile, int directLength,
        int reverbLength, int normSrc, int normMic, int normFreq, bool resetMic);
int harktoolcli_calctfgeo(int argc, char *argv[]);
int harktoolcli_calctfgeoExec(const char* tspFile, const char* micFile,
        enum TFType outputType, const char* outputFile, int directLength,
        int normSrc, int normMic, int normFreq, bool resetMic);
int harktoolcli_calctfmouth(int argc, char *argv[]);
int harktoolcli_calctfmouthExec(const char* tspFile, const char* micFile,
        enum TFType outputType, const char* outputFile, int directLength,
        int reverbLength, int normSrc, int normMic, int normFreq, bool resetMic,
        int tspEnd, int mch, int hop, int headMargin);
int harktoolcli_calcsysimp(int argc , char *argv[]);

int *harktool_newIntOneDim(int size_x);
void harktool_freeIntOneDim(int *matrix);
COMPLEX_FLOAT* harktool_newComplexFloatOneDim(int size_x);
void harktool_freeComplexFloatOneDim(COMPLEX_FLOAT* matrix);

char **harktool_newCharTwoDim(int size_x , int size_y);
void harktool_freeCharTwoDim(char **matrix);
float **harktool_newFloatTwoDim(int size_x , int size_y);
void harktool_freeFloatTwoDim(float** matrix);
COMPLEX_DOUBLE** harktool_newComplexDoubleTwoDim(int size_x, int size_y);
void harktool_freeComplexDoubleTwoDim(COMPLEX_DOUBLE** matrix);

float*** harktool_newFloatThreeDim(int size_x, int size_y, int size_z);
void harktool_freeFloatThreeDim(float*** matrix);
COMPLEX_FLOAT*** harktool_newComplexFloatThreeDim(int size_x, int size_y, int size_z);
void harktool_freeComplexFloatThreeDim(COMPLEX_FLOAT ***matrix);


void harktool_new(int type) ;
void harktool_delete(int type) ;
void harktool_initArray(int type) ;
void harktool_DoubleThreeDim(double ***matrix , int size_x, int size_y);
void harktool_setIRvectFromFlt(enum TFType outputType, int src, int directLength, int reverbLength, SRC_POS_INFO* srcPosInfo);
void harktool_setIRvectFromWav(enum TFType outputType, int directLength, int reverbLength, SRC_POS_INFO srcPosInfos[]);
void harktool_getWavDataFromFile(const char *fname, int* frameLength, int* samplingRate);
void harktool_getLocaltime(char *buff) ;
void harktool_createTmpPath() ;
int harktool_CheckDir(char *dirname) ;
int harktool_createDir(char* path) ;
int harktool_split(char *str, const char *delim, int *outlist)  ;
void harktool_getFileSize(char* filename  , long *size);
void harktool_setBufferFromFile(float *fltBuf , long file_size , char *fname) ;
void harktool_normalizeIR(float*** f3dvect) ;
void harktool_normalizeTF() ;
void normalizeTF(int normSrc, int normMic, int normFreq);
void harktool_normalizeSRC();
void harktool_normalizeMIC();
void harktool_normalizeFREQ();
void showReverbSources(int src, int peakTopIdx, const int* peaks);
void harktool_FFT(float*** f3dvect);
void harktool_conjTF();
int searchPeak(const float* irVect, int irLength, int start, int end);

void harktool_copyTmpTFvectFromTFvect() ;
void harktool_copyTmpIRvectToTMP(float*** f3dvect);
void harktool_absTF();
void harktool_powTF();
void harktool_powIR();
float harktool_sumAll(int src, int type,float*** f3dvect);
void harktool_divAll(int src, float sum , int type, float*** f3dvect);
//void harktool_writeTF() ;
void writeTF(enum TFType tfType, const char* output, bool resetMic);
void harktool_setLoctfFromArray();
void harktool_setSeptfFromArray();
void harktool_setZeroToTFvect() ;
void harktool_freeMemory() ;
void harktool_Exit(int exit_type) ;
void harktool_init() ;
void harktool_setMicPos() ;
void setMicPos(const int* MicPosIds, int Nmic, const harkio_Positions* poses, float** MicPosVect);
void harktool_setSrcPos() ;
void harktool_calcMicPosMean() ;
float harktool_pow(float a , int cnt) ;
void harktool_setMicPosId(const bool* outChannels);
//void setMicPosIds(const harkio_Positions* poses, int Nmic, int* MicPosIds);
void harktool_setSrcPosId();
void harktool_BubSort(int a[ ], int n) ;
void harktool_ShowData(int x[ ], int n);
void harktool_help(char* funcname) ;
void Log(int logType, const char *funcname, int lineNo,  char *format, ...);
void harktool_randomChar() ;
void harktool_swapFloatComplex(float _Complex  *a , float _Complex  *b);
void harktool_swapDoubleComplex(double _Complex  *a , double _Complex  *b);
//int harktool_calcDoubleComplex( double _Complex *value , double _Complex valuea , double _Complex valueb , enum Calculate calc);
void setFftData();
void initFftData(int Nfft, COMPLEX_FLOAT* buff_w);

void harktool_setChannelList();
COMPLEX_FLOAT COMPLEX_FLT(COMPLEX_FLOAT z);
COMPLEX_DOUBLE COMPLEX_DBL(double real , double imag);
//int harktool_calcSysImp(int argc , ...);
int harktool_calcSysImp(enum SoundType tspOrgType, const char* tspOrgFile, int tspOrgChannel, int signalMax, int cfgSamplingRate,
        int sampleByFrame, int syncAddFreq, int syncAddOffset, int searchOffsetFrom/* =-1 */, int searchOffsetTo/* =-1 */,
        const bool* outChannels, int chOrigin, const char* inputWaveFile, const char* outputFltFile);

//void createReadMeToZipFile(char *zip_path);
int appendInfoToZipFile(const char *zip_path, const SRC_POS_INFO* srcPosInfo);

void harktool_helpCalcSysImp();
void harktool_getArguments(int argc , char** argv);
void harktool_SetFloatArray(float *fltBuf , long file_size , char *fname);

int harktool_mouthTspFD(const char *inFile   , // Input wav file name [String]
                        const char *outFile  , // Output flt filename prefix. The name will be [outDir outFile '_ch*.flt']  [String]
                        int s                , // start point for Transfer Function(TF) calculation in sample [Integer]
                        int e                , // end point for TF calculation in sample [Integer]
                        int mch              , // Channel index of the voice input. [Integer]
                        const bool* outChannels, // Select channels for output.(MAX_CH)
                        int fltlen           , // Number of samples output as TF [Integer]
                        int nfft             , // Number of FFT points [Integer]
                        int hop              , // Shift length for TF calculation [Integer]
                        int headmargin       , // Head margin of impulse responces in samples [Integer(Default:0)]
                        bool enable_wn       , // Enables adding white noise to escape unstable zero signal [bool]
                        double wn_level      , // Level of white noise. Set enable_wn=true if you want to use this. [Double]
                        int cfgSamplingRate    // Sampling rate of configuration
                        );

long get_flt_size(const char* fname);
int read_flt(const char* fname, double* dBuf, size_t file_size, double* signal_max);
int write_flt(const char* fname, const double* dBuf, size_t count);
int readwavfile(const char *fname, double ***wave, int *length, int *ch, int *samplerate);
int writewavfile(char *fname, double **wave, int length, int ch, int samplerate);

void harktool_FFTImpl_fFFT(const COMPLEX_FLOAT *a, COMPLEX_FLOAT *A, int fftsize);
void harktool_FFTImpl_fFFTInv(const COMPLEX_FLOAT *A, COMPLEX_FLOAT *a, int fftsize);
void harktool_FFTImpl_fRealFFT(const float *float_a, COMPLEX_FLOAT *A, int fftsize);
void harktool_FFTImpl_fRealFFTInv(const COMPLEX_FLOAT  *A , float *float_a , int fftsize);
void harktool_FFTImpl_dFFT(const COMPLEX_DOUBLE *a, COMPLEX_DOUBLE *A, int fftsize);
void harktool_FFTImpl_dFFTInv(const COMPLEX_DOUBLE *A, COMPLEX_DOUBLE *a, int fftsize);
void harktool_FFTImpl_dRealFFT(const double *double_a, COMPLEX_DOUBLE  *A, int fftsize);
void harktool_FFTImpl_dRealFFTInv(const COMPLEX_DOUBLE  *A , double *double_a , int fftsize);

void genMUSIC();
void genGSS(int normSrc, int normMic, int normFreq);
void genTfFromIR(enum TFType outputType, int normSrc, int normMic, int normFreq, bool resetMic, const char* outputFile,
        const SRC_POS_INFO* srcPosInfo);

//int harktool_calcFloatComplex( float _Complex *value , float _Complex valuea , float _Complex valueb , enum Calculate calc);

void makeTempFltPath(int src, const harkio_Position* pos);

#ifdef WIN32
int vasprintf(char **strp, const char *format, va_list ap);
int asprintf(char **strp, const char *format, ...);
#endif

// *strp += asprintf()
int cat_asprintf(char** strp, const char* format, ...);

/**
 * zip内のファイルを読み込んで、必要な領域を*pStrBufに確保して格納する。
 * 読み込みに失敗したらEXIT_SUCCESS, 失敗したらEXIT_FAILUREを返す。
 * *pStrBufはfreeが必要
 */
int as_zip_fread(struct zip* handle, const char* srcPath, char** pStrBuf);
#endif
