#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <stdbool.h>
#include <popt.h>
#include <libharkio3.h>

#define TSP_ORG_FILE "16384.little_endian.wav"
#define TSP_ORG_PATH UTIL_PATH "/" TSP_ORG_FILE

bool hasParsedOpt = false;
void help(poptContext con, int exitCode) {
    if (hasParsedOpt)
        return;

    printf("\n");
    printf("\n");
    printf("NAME:\n");
    printf("  harktoolcli-genconfig\n");
    printf("\n");
    printf("Description:\n");
    printf("  The command creates a xml file that contians a config tag\n");
    printf("\n");
    poptPrintHelp(con, stdout, 0);
    printf("\n");
    printf("\n");
    exit(exitCode);
}

int main(int argc, char **argv) {
    harkio_XML *xml;
    FILE *fp;
    char* comment = "comment";
    int syncAvg = 16;
    int offset = 0;
    char* peakStr = "0,0";
    long peakFrom = 0, peakTo = 0;
    int nfft = 512;
    int samplingRate = 16000;
    int signalMax = 32768;
    int tspLength = 16384;
    char* tspPath = TSP_ORG_PATH;
    char* logLevel = "E";
    struct poptOption optionsTable[] = {
    /*{ "longname", "shortname",
     *  argInfo,
     *  *arg, retval,
     *  description, argment description} */
    { "comment", 'c',
      POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT,
      &comment, 'c',
      "Comment", NULL },
    { "sync", 's',
      POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &syncAvg, 's',
      "Number of TSPs in the recorded wav", NULL },
    { "offset", 'o',
      POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &offset, 'o',
      "Offset of the TSP", NULL },
    { "peak", 'p',
      POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT,
      &peakStr, 'p',
      "Range for peak search (1-based, 0 means all range)", "INT,INT" },
    { "nfft", 'n',
      POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &nfft, 'n',
      "NFFT of convolution", NULL },
    { "rate", 'r',
      POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &samplingRate, 'r',
      "SamplingRate", NULL },
    { "max", 'm',
      POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &signalMax, 'm',
      "SignalMax", NULL },
    { "length", 'l',
      POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &tspLength, 'l',
      "TSPlength", NULL },
    { "tsp-path", 't',
      POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT,
      &tspPath, 't',
      "Path to the original TSP", NULL },
    { "logLevel", '\0',
      POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT,
      &logLevel, '-',
      "Log information level", "E|W|I|D" },
    { "help", '\0',
      0,
      NULL, '?',
      "Show this help message", NULL },
    POPT_TABLEEND };

    poptContext optCon = poptGetContext(NULL, argc, (const char**)argv, optionsTable, 0);
    poptSetOtherOptionHelp(optCon, "[OPTIONS]* [output_filename]");

    // argc check
//    if (argc <= 1) {
//        harkio_Log_printf(harkio_Log_Error, "argument error!!");
//        help(optCon, EXIT_FAILURE);
//    }

    // option processing
    int c;
    while ((c = poptGetNextOpt(optCon)) >= 0) {
        switch (c) {
            char *tokenFrom, *tokenTo, *end;
        case 'p':
            tokenFrom = strtok(peakStr, ",");
            tokenTo = strtok(NULL, ",");
            if (tokenFrom && tokenTo) {
                peakFrom = strtol(tokenFrom, &end, 10);
                if (!(end && *end == '\0') || peakFrom > INT_MAX || peakFrom < 0) {
                    c = POPT_ERROR_BADNUMBER;
                    goto OPTIONERR;
                }
                peakTo = strtol(tokenTo, &end, 10);
                if (!(end && *end == '\0') || peakTo > INT_MAX || peakTo < 0) {
                    c = POPT_ERROR_BADNUMBER;
                    goto OPTIONERR;
                }
            } else {
                c = POPT_ERROR_BADNUMBER;
                goto OPTIONERR;
            }
            hasParsedOpt = true;
            break;
        case '-':
            if (strcmp(logLevel, "W") == 0) {
                harkio_Log_setLevel(harkio_Log_Warning);
            } else if (strcmp(logLevel, "I") == 0) {
                harkio_Log_setLevel(harkio_Log_Info);
            } else if (strcmp(logLevel, "D") == 0) {
                harkio_Log_setLevel(harkio_Log_Debug);
            } else {
                harkio_Log_setLevel(harkio_Log_Error);
            }
            hasParsedOpt = true;
            break;
        case '?':
            // help
            help(optCon, EXIT_SUCCESS);
            /* no break */
        default:
            hasParsedOpt = true;
        }
    }
    if (c < -1) {
        OPTIONERR:
        /* an error occurred during option processing */
        harkio_Log_printf(harkio_Log_Error, "%s: %s",
                poptBadOption(optCon, POPT_BADOPTION_NOALIAS), poptStrerror(c));
        exit(EXIT_FAILURE);
    }

    // get filename
    char* outfile = poptGetArg(optCon);

    harkio_Config *cfg = harkio_Config_new();

    // set values
    harkio_Config_setComment(cfg, comment);
    harkio_Config_setSynchronousAverage(cfg, syncAvg);
    harkio_Config_setTSPoffset(cfg, offset);
    harkio_Config_setPeakSearchFrom(cfg, (int)peakFrom);
    harkio_Config_setPeakSearchTo(cfg, (int)peakTo);
    harkio_Config_setNfft(cfg, nfft);
    harkio_Config_setTSPpath(cfg, tspPath);
    harkio_Config_setSamplingRate(cfg, samplingRate);
    harkio_Config_setSignalMax(cfg, signalMax);
    harkio_Config_setTSPlength(cfg, tspLength);

    // debug print
    harkio_Config_print(cfg);

    if (outfile == NULL) {
        fp = stdout;
    }else if ((fp = fopen(outfile, "w")) == NULL) {
        harkio_Log_FileOpenFailed(outfile);
        exit(EXIT_FAILURE);
    }

    // make new XML object and write
    xml = harkio_XML_newFromData(NULL, NULL, cfg);

    harkio_XML_write(xml, fp);

    harkio_XML_delete(&xml);

    if (outfile != NULL)
        fclose(fp);

    fprintf(stderr, "SUCCESS\n");
    if (outfile != NULL)
        fprintf(stderr, " output file : %s\n", outfile);

    // free poptContext
    poptFreeContext(optCon);

    exit(EXIT_SUCCESS);
}
